#If Not %Def( %INCLUDE )
#Compile SLL "Message.sll"
#Dim All
#Compiler PBWin 10.03
#EndIf

#Include Once "win32api.inc"

Function NewMessageTrack() Common Export As iMessageTrack
	Local obj As iMessageTrack
	obj = Class "cMessageTrack"
	Function = obj
End Function

' Consider replacing this implementation with a dictionary of DWord, iMessage
' it will be slower but it will also be safer
Class cMessageTrack Guid$("{3c255bde-47a5-48d6-bcdb-c18815782418}") Common
	Instance objs_() As iMessage
	Instance cnt_ As Long

	Class Method Create()
	End Method
	Class Method Destroy()
		Local i As Long
		For i=0 To cnt_
			objs_(i) = Nothing
		Next
		Erase objs_()
	End Method

	Interface iMessageTrack Guid$("{b1078b6f-ee69-4464-bd1f-cdc5dc6c282a}")
		Inherit IUnknown

		Method Add( obj As iMessage )
			Local i As Long
			Local ub As Long

			If IsFalse IsInterface( obj, iMessage ) Then Exit Method
			If IsWindow( obj.HWnd ) = 0 Then Exit Method

			For i=0 To cnt_-1
				If IsFalse IsObject( objs_(i) ) Then Exit For
			Next
			ub = UBound( objs_() )
			If i = cnt_ And cnt_ >= ub Then
				ReDim Preserve objs_( ub+10 )
			End If
			objs_(i) = obj
			SetWindowLong( obj.HWnd, %GWL_USERDATA, i )
			If i=cnt_ Then Incr cnt_
		End Method

		Method Remove( obj As iMessage )
			Local idx As Long
			If IsFalse IsInterface( obj, iMessage ) Then Exit Method
			idx = GetWindowLong( obj.HWnd, %GWL_USERDATA )
			If idx>=0 And idx<cnt_ Then objs_(idx) = Nothing
		End Method

		Property Get Item( ByVal hWnd As Dword ) As iMessage
			Local idx As Long
			idx = GetWindowLong( hWnd, %GWL_USERDATA )
			If idx>=0 And idx<cnt_ Then
				Property = objs_(idx)
			End If
		End Property
	End Interface
End Class

Global gbrTrack As iMessageTrack

Function NewMessageBridge(bridgeTo As iMessage) Common Export As iMessageBridge
	Local obj As iMessageBridge
	obj = Class "cMessageBridge"
	obj.SetBridge( bridgeTo )
	Function = obj
End Function

Class cMessageBridge Guid$("{4c07094b-1bea-4150-85ce-273934df9e22}") Common
	Instance bridgeTo_ As iMessage
	Instance defProc As Dword

	Class Method Create()
		' Ensure class tracking conversion from functions to classes exists
		If IsFalse IsInterface( gbrTrack, iMessageTrack ) Then gbrTrack = NewMessageTrack()
	End Method

	Class Method Destroy()
		Dim bridge As iMessageBridge
		bridge = Me
		bridge.ClearBridge()
	End Method

	Interface iMessageBridge Guid$("{3fa159d7-c691-4d6c-9104-5e034469097b}")
		Inherit IUnknown

		Method SetBridge( bridgeTo As iMessage )
			Local bridge As iMessage
			Local mb As iMessageBridge

			mb = Me

			Call mb.ClearBridge()

			If IsFalse IsInterface( bridgeTo, iMessage ) Then Exit Method

			' Sub-class if window didn't use our control proc
			defproc = GetWindowLong( bridgeTo.HWnd, %GWL_WNDPROC)
			If defproc <> CodePtr(Bridge_ControlProc) Then
				 SetWindowLong(bridgeTo.HWnd, %GWL_WNDPROC, CodePtr(Bridge_ControlProc) )
			End If

			bridgeTo_ = bridgeTo
			bridge = Me
			gbrTrack.Add( bridge )
		End Method
		Method ClearBridge()
			Local bridge As iMessage

			' Reverse subclassing
			If IsInterface( bridgeTo_, iMessage ) Then
				If defproc And defproc <> CodePtr( Bridge_ControlProc ) Then
					 SetWindowLong(bridgeTo_.HWnd, %GWL_WNDPROC, defProc )
				End If

				bridge = Me
				gbrTrack.Remove( bridge )
				bridgeTo_ = Nothing
			End If
		End Method
	End Interface

	' Process messages
	Interface iMessage Guid$("{2e0e23cf-6e19-4a88-a33c-2cca8397778f}")
		Inherit IUnknown

		Property Get HWnd() As Long
			Property = bridgeTo_.HWnd
		End Property
		Method Proc( ByVal wMsg As Dword, ByVal wParam As Long, ByVal lParam As Long, ByRef ret As Long) As Long
			If IsInterface( bridgeTo_, iMessage ) Then
				If IsFalse bridgeTo_.Proc( wMsg, wParam, lParam, ret ) Then
					If defProc And defProc <> CodePtr( Bridge_ControlProc ) Then
						ret = CallWindowProc( defProc, bridgeTo_.HWnd, wMsg, wParam, lParam )
					Else
						ret = DefWindowProc( bridgeTo_.HWnd, wMsg, wParam, lParam )
					End If
					Method = -1
				End If
			End If
		End Method
	End Interface
End Class

'==============================================================
' Bridge to class
'==============================================================
Function Bridge_ControlProc_Address() Export Common As Dword
	Function = CodePtr( Bridge_ControlProc )
End Function

Function Bridge_ControlProc( ByVal hWnd As Dword, ByVal wMsg As Dword, ByVal wParam As Long, ByVal lParam As Long) As Long
	Local obj As iMessage
	Local ret As Long

	If IsFalse IsObject( gbrTrack ) Then Exit Function
	obj = gbrTrack.Item( hWnd )
	If IsInterface( obj, iMessage ) Then
		If obj.Proc( wMsg, wParam, lParam, ret ) Then
			Function = ret
			Exit Function
		End If
	End If
	Function = DefWindowProc( hWnd, wMsg, wParam, lParam )
End Function
